//  cwMOEAD_main.java
//
//  This file is based on the MOEAD.java from the jMetal library. The copyright information for the original file
//  follows.
//
//  Author: blinded
//
//  MOEAD_main.java
//
//  Author:
//       Antonio J. Nebro <antonio@lcc.uma.es>
//       Juan J. Durillo <durillo@lcc.uma.es>
//
//  Copyright (c) 2011 Antonio J. Nebro, Juan J. Durillo
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
// 
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

package gecco2015.cwmoead;

import gecco2015.utils.Hypervolume;
import gecco2015.utils.LogItem;
import jmetal.core.Algorithm;
import jmetal.core.Operator;
import jmetal.core.Problem;
import jmetal.core.SolutionSet;
import jmetal.operators.crossover.CrossoverFactory;
import jmetal.operators.mutation.MutationFactory;
import jmetal.problems.WFG.*;
import jmetal.problems.ZDT.*;
import jmetal.qualityIndicator.QualityIndicator;
import jmetal.util.Configuration;
import jmetal.util.JMException;
import jmetal.util.comparators.ObjectiveComparator;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.FileHandler;
import java.util.logging.Logger;

/**
 * This class executes the algorithm described in:
 * H. Li and Q. Zhang,
 * "Multiobjective Optimization Problems with Complicated Pareto Sets,  MOEA/D
 * and NSGA-II". IEEE Trans on Evolutionary Computation, vol. 12,  no 2,
 * pp 284-302, April/2009.
 */
public class cwMOEAD_main {
    public static Logger logger_;      // Logger object
    public static FileHandler fileHandler_; // FileHandler object

    /**
     * @param args Command line arguments. The first (optional) argument specifies
     *             the problem to solve.
     * @throws jmetal.util.JMException
     * @throws java.io.IOException
     * @throws SecurityException       Usage: three options
     *                                 - jmetal.metaheuristics.moead.MOEAD_main
     *                                 - jmetal.metaheuristics.moead.MOEAD_main problemName
     *                                 - jmetal.metaheuristics.moead.MOEAD_main problemName ParetoFrontFile
     * @throws ClassNotFoundException
     */
    public static void main(String[] args) throws JMException, SecurityException, IOException, ClassNotFoundException {
        Problem problem;         // The problem to solve
        Algorithm algorithm;         // The algorithm to use
        Operator crossover;         // Crossover operator
        Operator mutation;         // Mutation operator

        QualityIndicator indicators; // Object to get quality indicators

        HashMap parameters; // Operator parameters

        // Logger object and file to store log messages
        logger_ = Configuration.logger_;

        Problem[] problems = {new ZDT1("Real"), new ZDT2("Real"), new ZDT3("Real"), new ZDT4("Real"), new ZDT6("Real"),
                new WFG1("Real", 4, 20, 2),
                new WFG2("Real", 4, 20, 2),
                new WFG3("Real", 4, 20, 2),
                new WFG4("Real", 4, 20, 2), new WFG5("Real", 4, 20, 2), new WFG6("Real", 4, 20, 2),
                new WFG7("Real", 4, 20, 2), new WFG8("Real", 4, 20, 2), new WFG9("Real", 4, 20, 2)
        };


        for (int problemID = 0; problemID < problems.length; problemID++) {
            problem = problems[problemID];
            int repeats = 15;

            for (int run = 0; run < repeats; run++) {

                algorithm = new cwMOEAD(problem);
                //algorithm = new MOEAD_DRA(problem);

                Hypervolume hv = new Hypervolume(problem);
                indicators = new QualityIndicator(problem, "fronts/" + problem.getName() + ".pf");

                // Add the indicator object to the algorithm
                algorithm.setInputParameter("indicators", indicators);

                // Algorithm parameters
                algorithm.setInputParameter("populationSize", 100);
                algorithm.setInputParameter("maxEvaluations", 150000);

                // Directory with the files containing the weight vectors used in
                // Q. Zhang,  W. Liu,  and H Li, The Performance of a New Version of MOEA/D
                // on CEC09 Unconstrained MOP Test Instances Working Report CES-491, School
                // of CS & EE, University of Essex, 02/2009.
                // http://dces.essex.ac.uk/staff/qzhang/MOEAcompetition/CEC09final/code/ZhangMOEADcode/moead0305.rar
                algorithm.setInputParameter("dataDirectory", "moeadWeights");

                algorithm.setInputParameter("finalSize", 300); // used by MOEAD_DRA

                algorithm.setInputParameter("T", 20);
                algorithm.setInputParameter("delta", 0.9);
                algorithm.setInputParameter("nr", 2);

                // Crossover operator
                parameters = new HashMap();
                parameters.put("CR", 1.0);
                parameters.put("F", 0.5);
                crossover = CrossoverFactory.getCrossoverOperator("DifferentialEvolutionCrossover", parameters);

                // Mutation operator
                parameters = new HashMap();
                parameters.put("probability", 1.0 / problem.getNumberOfVariables());
                parameters.put("distributionIndex", 20.0);
                mutation = MutationFactory.getMutationOperator("PolynomialMutation", parameters);

                algorithm.addOperator("crossover", crossover);
                algorithm.addOperator("mutation", mutation);

                // Execute the Algorithm
                long initTime = System.currentTimeMillis();
                SolutionSet population = algorithm.execute();
                long estimatedTime = System.currentTimeMillis() - initTime;

                // Result messages
                logger_.info("Total execution time: " + estimatedTime + "ms");

                ArrayList<LogItem> log = (ArrayList<LogItem>) algorithm.getOutputParameter("indicatorsLog");

                String outputDir = "output/" + problem.getName() + "/cwMOEAD/";
                new File(outputDir).mkdirs();

                printLog(log, outputDir + "indicators." + run, hv.getOptimalHypervolume());
                population.sort(new ObjectiveComparator(0));
                population.printVariablesToFile(outputDir + "variables." + run);
                population.printObjectivesToFile(outputDir + "objectives." + run);


                if (indicators != null) {
                    logger_.info("Quality indicators");
                    logger_.info("Hypervolume: " + indicators.getHypervolume(population));
                    logger_.info("EPSILON    : " + indicators.getEpsilon(population));
                    logger_.info("GD         : " + indicators.getGD(population));
                    logger_.info("IGD        : " + indicators.getIGD(population));
                    logger_.info("Spread     : " + indicators.getSpread(population));
                } // if
            }
        }
    } //main

    public static void printLog(ArrayList<LogItem> log, String fileName, double optimalHV) {

        PrintWriter pw = null;
        try {
            pw = new PrintWriter(new BufferedWriter(new FileWriter(fileName)));
        } catch (IOException e) {
            e.printStackTrace();
        }

        pw.println("#Evals\tdeltaH\tHyperV\tEps\tIGD");

        for (LogItem li : log) {
            pw.println(li.evaluations + "\t" + (optimalHV - li.hv) + "\t" + li.hv + "\t" + li.eps + "\t" + li.igd);
        }

        pw.close();
    }


} // MOEAD_main
