package gecco2015.utils;

import jmetal.core.Problem;
import jmetal.core.Solution;
import jmetal.core.SolutionSet;
import jmetal.util.comparators.ObjectiveComparator;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

public strictfp class Hypervolume {

    Problem prob;
    double xr;
    double yr;
    double optimalHypervolume;

    public Hypervolume(Problem prob) {
        this.prob = prob;
        setReference(prob);
        computeOptimalHypervolume();
    }

    private void setReference(Problem prob) {

        if (prob.getName().startsWith("ZDT")) {
            xr = 11.0;
            yr = 11.0;
        }

        if (prob.getName().startsWith("WFG")) {
            xr = 3.0;
            yr = 5.0;
        }

    }

    private void computeOptimalHypervolume() {

        String fileName = "fronts/" + prob.getName() + ".pf";

        ArrayList<Solution> sols = new ArrayList<>();

        try {
            BufferedReader br = new BufferedReader(new FileReader(fileName));
            String line = br.readLine();
            while (line != null) {
                Solution s = new Solution(prob);
                String[] nums = line.split("[ \t]");
                s.setObjective(0, Double.parseDouble(nums[0]));
                s.setObjective(1, Double.parseDouble(nums[1]));
                sols.add(s);
                line = br.readLine();
            }

            SolutionSet optim = new SolutionSet(sols.size());
            for (int i = 0; i < sols.size(); i++) {
                optim.add(sols.get(i));
            }

            optimalHypervolume = getHypervolume(optim);

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }

    }

    public double getOptimalHypervolume() {
        return optimalHypervolume;
    }

    public double getXr() {
        return xr;
    }

    public double getYr() {
        return yr;
    }

    public double getHypervolume(SolutionSet sols) {

        StrictRanking rank = new StrictRanking(sols);
        SolutionSet tmp = rank.getSubfront(0);

        tmp.sort(new ObjectiveComparator(0));

        double hypervolume = 0.0;
        double prevy = yr;
        for (int i = 0; i < tmp.size(); i++) {
            if (tmp.get(i).getObjective(1) > yr)
                continue;
            if (tmp.get(i).getObjective(0) > xr) {
                continue;
            }
            hypervolume += (prevy-tmp.get(i).getObjective(1))*(xr - tmp.get(i).getObjective(0));
            prevy = tmp.get(i).getObjective(1);
        }

        return hypervolume;
    }

    public void hypervolumeContributionAssignment(SolutionSet solutionSet, int nObjs) {
        int size = solutionSet.size();

        if (nObjs != 2) {
            throw new RuntimeException("Hypervolume contribution assignment implemented only for nObjs=2");
        }

        if (size == 0)
            return;

        if (size == 1) {
            solutionSet.get(0).setCrowdingDistance(Double.POSITIVE_INFINITY);
            return;
        } // if

        if (size == 2) {
            solutionSet.get(0).setCrowdingDistance(Double.POSITIVE_INFINITY);
            solutionSet.get(1).setCrowdingDistance(Double.POSITIVE_INFINITY);
            return;
        } // if

        //Use a new SolutionSet to evite alter original solutionSet
        SolutionSet front = new SolutionSet(size);
        for (int i = 0; i < size; i++){
            front.add(solutionSet.get(i));
        }

        for (int i = 0; i < size; i++)
            front.get(i).setCrowdingDistance(0.0);

        double distance;

        // Sort the population by Obj n
        front.sort(new ObjectiveComparator(0));

        //Set de crowding distance
        front.get(0).setCrowdingDistance(Double.POSITIVE_INFINITY);
        front.get(size-1).setCrowdingDistance(Double.POSITIVE_INFINITY);

        while (front.size() > 2) {
            double minContribution = Double.MAX_VALUE;
            int minIdx = 0;
            for (int j = 1; j < front.size()-1; j++) {
                int prev = j-1;
                int next = j+1;

                double xN = front.get(next).getObjective(0);
                double xC = front.get(j).getObjective(0);
                double yC = front.get(j).getObjective(1);
                double yP = front.get(prev).getObjective(1);

                double contrib = (xN-xC)*(yP-yC);

                if (contrib < minContribution) {
                    minContribution = contrib;
                    minIdx = j;
                }

                front.get(j).setCrowdingDistance(contrib);
            }
            front.remove(minIdx);
        }
    } // crowdingDistanceAssing






}





