package gecco2015.utils;

import jmetal.core.Solution;
import jmetal.core.SolutionSet;
import jmetal.util.comparators.DominanceComparator;
import jmetal.util.comparators.OverallConstraintViolationComparator;

import java.util.ArrayList;
import java.util.Comparator;

/**
 * This class implements some facilities for ranking solutions.
 * Given a <code>SolutionSet</code> object, their solutions are ranked 
 * according to scheme proposed in NSGA-II; as a result, a set of subsets 
 * are obtained. The subsets are numbered starting from 0 (in NSGA-II, the 
 * numbering starts from 1); thus, subset 0 contains the non-dominated 
 * solutions, subset 1 contains the non-dominated solutions after removing those
 * belonging to subset 0, and so on.
 */
public class StrictRanking {

    /**
     * The <code>SolutionSet</code> to rank
     */
    private SolutionSet   solutionSet_ ;

    /**
     * An array containing all the fronts found during the search
     */
    private SolutionSet[] ranking_  ;
    private static ArrayList<Integer> lastFrontIdx;
    private ArrayList<ArrayList<Integer>> subfrontIdxs;
    private static ArrayList<Integer> indx;

    /**
     * stores a <code>Comparator</code> for dominance checking
     */
    private static final Comparator dominance_ = new DominanceComparator();

    /**
     * stores a <code>Comparator</code> for Overal Constraint Violation Comparator
     * checking
     */
    private static final Comparator constraint_ = new OverallConstraintViolationComparator();

    /**
     * Constructor.
     * @param solutionSet The <code>SolutionSet</code> to be ranked.
     */
    public StrictRanking(SolutionSet solutionSet) {
        solutionSet_ = solutionSet ;

        ArrayList<Solution> sols = new ArrayList<>();
        indx = new ArrayList<>();


        for (int i = 0; i < solutionSet.size(); i++) {
            sols.add(solutionSet.get(i));
            indx.add(i);
        }

        ArrayList<ArrayList<Solution>> fronts = new ArrayList<>();
        ArrayList<ArrayList<Integer>> frontsIdx = new ArrayList<>();

        while (sols.size() > 0) {
            ArrayList<Solution> front = getFirstFront(sols);
            fronts.add(front);
            frontsIdx.add(lastFrontIdx);
            sols.removeAll(front);
            indx.removeAll(lastFrontIdx);
        }

        ranking_ = new SolutionSet[fronts.size()];

        for (int i = 0; i < ranking_.length; i++) {
            ranking_[i] = new SolutionSet(fronts.get(i).size());
            for (int j = 0; j < fronts.get(i).size(); j++) {
                ranking_[i].add(fronts.get(i).get(j));
            }
        }

        subfrontIdxs = frontsIdx;

    }

    public static boolean dominates(Solution a, Solution b) {
        int dom_count = 0;
        for (int i = 0; i < a.getNumberOfObjectives(); i++) {
            if (a.getObjective(i) <= b.getObjective(i))
                dom_count++;
        }
        if (dom_count == a.getNumberOfObjectives())
            return true;
        return false;
}

    public static ArrayList<Solution> getFirstFront(ArrayList<Solution> solutions) {
        ArrayList<Solution> front = new ArrayList<>();
        ArrayList<Integer> frontIdx = new ArrayList<>();
        front.add(solutions.get(0));
        frontIdx.add(indx.get(0));
        for (int pi = 0; pi < solutions.size(); pi++) {
            Solution p = solutions.get(pi);
            if (front.contains(p))
                continue;
            front.add(p);
            frontIdx.add(indx.get(pi));
            ArrayList<Solution> toRemove = new ArrayList<>();
            ArrayList<Integer> toRemoveIdx = new ArrayList<>();
            for (int qi = 0; qi < front.size(); qi++) {
                Solution q = front.get(qi);
                int qIdx = frontIdx.get(qi);
                if (q.equals(p))
                    continue;
                if (toRemove.contains(q))
                    continue;
                if (dominates(q, p)) {
                    toRemove.add(p);
                    toRemoveIdx.add(indx.get(pi));
                }
                else if (dominates(p, q)) {
                    toRemove.add(q);
                    toRemoveIdx.add(qIdx);
                }
            }
            front.removeAll(toRemove);
            frontIdx.removeAll(toRemoveIdx);
        }

        lastFrontIdx = frontIdx;
        return front;
    }

    /**
     * Returns a <code>SolutionSet</code> containing the solutions of a given rank.
     * @param rank The rank
     * @return Object representing the <code>SolutionSet</code>.
     */
    public SolutionSet getSubfront(int rank) {
        return ranking_[rank];
    } // getSubFront

    public ArrayList<Integer> getSubfrontIdxs(int rank) {
        return subfrontIdxs.get(rank);
    }

    /**
     * Returns the total number of subFronts founds.
     */
    public int getNumberOfSubfronts() {
        return ranking_.length;
    } // getNumberOfSubfronts
} // Ranking
