//  FastHypervolume.java
//
//  Authors:
//       Antonio J. Nebro <antonio@lcc.uma.es>
//       Juan J. Durillo <durillo@lcc.uma.es>
//
//  Copyright (c) 2013 Antonio J. Nebro, Juan J. Durillo
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

package gecco2015.utils.fastHypervolume;

import gecco2015.utils.StrictRanking;
import gecco2015.utils.fastHypervolume.wfg.Front;
import gecco2015.utils.fastHypervolume.wfg.WFGHV;
import jmetal.core.Problem;
import jmetal.core.Solution;
import jmetal.core.SolutionSet;
import jmetal.util.comparators.CrowdingComparator;
import jmetal.util.comparators.ObjectiveComparator;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Created with IntelliJ IDEA.
 * User: Antonio J. Nebro
 * Date: 26/08/13
 * Time: 10:20
 */
public class FastHypervolume {
  Solution referencePoint_ ;
  int numberOfObjectives_ ;
  double offset_ = 20.0 ;
  boolean externalReference = false;
  Solution externalPoint;

  public FastHypervolume() {
    referencePoint_ = null ;
    numberOfObjectives_ = 0 ;
  }

  public FastHypervolume(double offset) {
    referencePoint_ = null ;
    numberOfObjectives_ = 0 ;
    offset_ = offset ;
  }

  public void setReferencePoint(Solution referencePoint) {
    this.referencePoint_ = referencePoint;
    this.externalPoint = referencePoint;
    this.externalReference = true;
  }

  public Solution getReferencePoint() {
    return referencePoint_;
  }

  public void setReferencePointByProblem(Problem prob) {
    externalReference = true;
    if (prob.getName().startsWith("ZDT")) {
      referencePoint_ =  new Solution(2);
      referencePoint_.setObjective(0, 11);
      referencePoint_.setObjective(1, 11);
    }
    if (prob.getName().startsWith("WFG")) {
      referencePoint_ = new Solution(prob.getNumberOfObjectives());
      for (int i = 0; i < prob.getNumberOfObjectives(); i++) {
        referencePoint_.setObjective(i, 2*i+3);
      }
    }
    this.externalPoint = referencePoint_;
  }

  public double computeOptimalHypervolume(Problem prob) {

    String fileName;

    if (prob.getNumberOfObjectives() == 2)
      fileName = "fronts/" + prob.getName() + ".pf";
    else {
      fileName =  "fronts/" + prob.getName() + "." + prob.getNumberOfObjectives() + "D.pf";
    }

    ArrayList<Solution> sols = new ArrayList<>();

    try {
      BufferedReader br = new BufferedReader(new FileReader(fileName));
      String line = br.readLine();
      while (line != null) {
        Solution s = new Solution(prob);
        String[] nums = line.split("[ \t]");
        for (int i = 0; i < prob.getNumberOfObjectives(); i++) {
          s.setObjective(i, Double.parseDouble(nums[i]));
        }
        sols.add(s);
        line = br.readLine();
      }

      SolutionSet optim = new SolutionSet(sols.size());
      for (int i = 0; i < sols.size(); i++) {
        optim.add(sols.get(i));
      }

      StrictRanking sr = new StrictRanking(optim);

      return computeHypervolume(sr.getSubfront(0));

    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }

    return Double.NaN;

  }

  public double computeHypervolume(SolutionSet solutionSet) {
    double hv ;
    if (solutionSet.size() == 0)
      hv = 0.0 ;
    else {
      numberOfObjectives_ = solutionSet.get(0).getNumberOfObjectives() ;
      referencePoint_ = new Solution(numberOfObjectives_) ;
      updateReferencePoint(solutionSet);
      if (numberOfObjectives_ == 2) {

        solutionSet.sort(new ObjectiveComparator(numberOfObjectives_-1, true));
        hv = get2DHV(solutionSet) ;
      }
      else {
        updateReferencePoint(solutionSet);
        Front front = new Front(solutionSet.size(), numberOfObjectives_, solutionSet) ;
        hv = new WFGHV(numberOfObjectives_, solutionSet.size(), referencePoint_).getHV(front) ;
      }
    }

    return hv ;
  }

  public double computeHypervolume(SolutionSet solutionSet, Solution referencePoint) {
    double hv = 0.0;
    if (solutionSet.size() == 0)
      hv = 0.0;
    else {
      numberOfObjectives_ = solutionSet.get(0).getNumberOfObjectives();
      referencePoint_ = referencePoint;

      if (numberOfObjectives_ == 2) {
        solutionSet.sort(new ObjectiveComparator(numberOfObjectives_ - 1, true));

        hv = get2DHV(solutionSet);
      } else {
        WFGHV wfg = new WFGHV(numberOfObjectives_, solutionSet.size());
        Front front = new Front(solutionSet.size(), numberOfObjectives_, solutionSet);
        hv = wfg.getHV(front, referencePoint);
      }
    }

    return hv;
  }


  /**
   * Updates the reference point
   */
  private void updateReferencePoint(SolutionSet solutionSet) {
    if (externalReference) {
      referencePoint_ = externalPoint;
      return;
    }
    double [] maxObjectives = new double[numberOfObjectives_] ;
    for (int i = 0; i < numberOfObjectives_; i++)
      maxObjectives[i] = 0 ;

    for (int i = 0; i < solutionSet.size(); i++)
      for (int j = 0 ; j < numberOfObjectives_; j++)
        if (maxObjectives[j] < solutionSet.get(i).getObjective(j))
          maxObjectives[j] = solutionSet.get(i).getObjective(j) ;

    for (int i = 0; i < referencePoint_.getNumberOfObjectives(); i++) {
      referencePoint_.setObjective(i, maxObjectives[i]+ offset_) ;
    }
  }

  /**
   * Computes the HV of a solution set.
   * REQUIRES: The problem is bi-objective
   * REQUIRES: The archive is ordered in descending order by the second objective
   * @return
   */
  public double get2DHV(SolutionSet solutionSet) {
    double hv = 0.0;
    if (solutionSet.size() > 0) {
      hv = Math.abs((solutionSet.get(0).getObjective(0) - referencePoint_.getObjective(0)) *
              (solutionSet.get(0).getObjective(1) - referencePoint_.getObjective(1)));

      for (int i = 1; i < solutionSet.size(); i++) {
        double tmp = Math.abs((solutionSet.get(i).getObjective(0) - referencePoint_.getObjective(0)) *
                (solutionSet.get(i).getObjective(1) - solutionSet.get(i - 1).getObjective(1)));
        hv += tmp;
      }
    }
    return hv;
  }

  /**
   * Computes the HV contribution of the solutions
   * @return
   */
  public void computeHVContributions(SolutionSet solutionSet) {
    double[] contributions = new double[solutionSet.size()] ;
    double solutionSetHV = 0 ;

    solutionSetHV = computeHypervolume(solutionSet) ;

    for (int i = 0; i < solutionSet.size(); i++) {
      Solution currentPoint = solutionSet.get(i);
      solutionSet.remove(i) ;

      if (numberOfObjectives_ == 2) {
        //updateReferencePoint(solutionSet);
        //solutionSet.sort(new ObjectiveComparator(numberOfObjectives_-1, true));
        contributions[i] = solutionSetHV - get2DHV(solutionSet) ;
      }
      else {
        Front front = new Front(solutionSet.size(), numberOfObjectives_, solutionSet) ;
        double hv = new WFGHV(numberOfObjectives_, solutionSet.size(), referencePoint_).getHV(front) ;
        contributions[i] = solutionSetHV - hv ;
      }
      solutionSet.add(i, currentPoint) ;
    }

    for (int i = 0; i < solutionSet.size(); i++) {
      solutionSet.get(i).setCrowdingDistance(contributions[i]) ;
    }
  }

  public void computeHVContributionsRemoving(SolutionSet sols) {

    if (sols.size() <= sols.get(0).getNumberOfObjectives()) {
      for (int i = 0; i < sols.size(); i++) {
        sols.get(i).setCrowdingDistance(0.0);  //all points are boundary points, the value does not matter
      }
      return;
    }

    SolutionSet tmpSols = new SolutionSet(sols.size());
    for (int i = 0; i < sols.size(); i++) {
      tmpSols.add(sols.get(i));
    }

    while (tmpSols.size() > sols.get(0).getNumberOfObjectives() + 1) {
      computeHVContributions(tmpSols);
      int idx = tmpSols.indexWorst(new CrowdingComparator());
      tmpSols.remove(idx);
    }
  }
}
